##
## khSims.R
## 31 august 2013
## 

require(e1071)
require(prob)

##
## run a simulation for the setting described in Kirby (2013)
##
## arguments
## _________
## (defaults are shown in the function definition)
## 
## nGens: Number of generations
## 
## nExamples: Number of examples received by each learner
##
## nLearners: Number of learners per generation
## 
## rBias: percentage of biased training examples
##
## cGroup: compensation for coarticulation (comp) or not (non)
## 
## storeIvl: store estimated PDF of p every storeIvl generations;
##           (not currently implemented)
##
## noise: should additional production noise be added? 
##        (not currently implemented)
##         
## writeFile: write the data out as a file?
##
## fPrefix: directory to write the file to (NB: need the trailing slash)
## 

setwd('/Users/jkirby/Dropbox/Documents/Khmer/simulation/labphon/')
source('khUtils.R')
source('khPlots.R')

khmerSim <- function(nGens=100, nExamples = 100, nLearners = 10,
                   rBias = 0.2, cGroup = 'comp', storeIvl = 5, noise = FALSE,
                   writeFile = TRUE, fPrefix = 'outfiles/'){

    stopifnot(is.numeric(nLearners) && round(nLearners)==nLearners && nLearners>0,
            is.numeric(nExamples) && round(nExamples)==nExamples && nExamples>0,
            is.numeric(nGens) && round(nGens)==nGens && nGens>0,
            is.numeric(storeIvl) && round(storeIvl)==storeIvl && storeIvl>0,
            is.logical(noise),
            is.logical(writeFile),
            is.character(fPrefix)
            )

    pre_est <- read.csv('pre-est.txt', header=T)

    teachers = data.frame()

    for (i in 1:nLearners) {
        teacher <- init(pre_est, nExamples)
        teacher$id <- rep(i, nExamples)
        teachers <- rbind(teachers, teacher)
    }

    rcutoff = mean(droplevels(subset(teachers, class=='kr'))$r) - (sd(droplevels(subset(teachers, class=='kr'))$r)*2)

    ## initial error
    pred <- naiveBayes(class~vot+f0+h1h2+r, data=teachers)
    eg<-data.frame(t=0,error=err(table(predict(pred, teachers), teachers[,5])))

    ## which timesteps to store p density at
    ## (not currently used)
    storeIts <- seq(from=storeIvl, to=nGens, by=storeIvl)
    if(!(nGens %in% storeIts)){ storeIts <- c(storeIts, n) }

    ## data frame to store densities
    densityDf <- data.frame(t=0, teachers)

    ## data frame to store means and sds 
    summaryStatsDf <- rbind(
        data.frame(t=0, cue='VOT (ms)', ddply(teachers, .(class), summarise, mean=mean(vot), sd=sd(vot))),
        data.frame(t=0, cue='f0 (Hz)', ddply(teachers, .(class), summarise, mean=mean(f0), sd=sd(f0))), 
        data.frame(t=0, cue='H1-H2 (dB)', ddply(teachers, .(class), summarise, mean=mean(h1h2), sd=sd(h1h2))),
        data.frame(t=0, cue='duration of /r/ (ms)', ddply(teachers, .(class), summarise, mean=mean(r), sd=sd(r)))
    )

    ## data frame to store simulation parameters
    simParamsDf <- data.frame(nGens = nGens, nExamples = nExamples, nLearners = nLearners, 
                              rBias = rBias, cGroup = cGroup, noise=noise, storeIvl = storeIvl)
    
    ## data frame to store error rate
    pred <- naiveBayes(class~vot+f0+h1h2+r, data=teachers)
    errorDf <- data.frame(t=0,error=err(table(predict(pred, teachers), teachers[,5])))

    progBar <- txtProgressBar(min = 0, max = nGens , style = 3)

    for (i in 1:nGens) {

        learners <- data.frame()

        for (j in 1:nLearners) {
        
            ## generate sample
            training <- sampleMVT(teachers, nExamples, nExamples, nExamples, noise=noise)

            ## bias
            kuu <- droplevels(subset(training, class=='k'))
            khuu <- droplevels(subset(training, class=='kh'))
            kruu <- droplevels(subset(training, class=='kr'))

            biased <- sample.df(kruu, nrow(kruu) * rBias)
            normal <- setdiff(kruu, biased) 
            rdeg = runif(1, min=0.5) ## rho
            #rdeg = 0.99

            biased$r = abs(biased$r - (biased$r*rdeg))
            if (cGroup=='non') biased$vot = biased$vot + (biased$r*(1*rdeg))
            kruu = rbind(biased,normal)
            training <- rbind(kuu, khuu, kruu)
                    
            ## learner classifies this sample
            ## if /r/ still means something
            if (mean(kruu$r) > rcutoff) { pred <- naiveBayes(class~vot+f0+h1h2+r, data=training) } 
            ## otherwise, stop paying attention to it
            else { pred <- naiveBayes(class~vot+f0+h1h2, data=training) }
            labels <- predict(pred, training)

            estimate <- training
            estimate$class <- labels
            estimate$id <- rep(j, nrow(estimate))

            learners <- rbind(learners, estimate)
        }

        ## store densities
        densityDf <- rbind(densityDf, data.frame(t=i, learners))

        ## data frame to store means and sds 
        summaryStatsDf <- rbind(summaryStatsDf,
            data.frame(t=i, cue='VOT (ms)', ddply(learners, .(class), summarise, mean=mean(vot), sd=sd(vot))),
            data.frame(t=i, cue='f0 (Hz)', ddply(learners, .(class), summarise, mean=mean(f0), sd=sd(f0))),
            data.frame(t=i, cue='H1-H2 (dB)', ddply(learners, .(class), summarise, mean=mean(h1h2), sd=sd(h1h2))),
            data.frame(t=i, cue='duration of /r/ (ms)', ddply(learners, .(class), summarise, mean=mean(r), sd=sd(r)))
        )

        classErr <- err(table(learners[,5], teachers[,5]))
        #cat('\nEpoch', i, 'error (predicted vs. intended): ', classErr)
        errorDf <- rbind(errorDf, c(i, classErr))
        teachers <- learners

        ## update progress bar
        setTxtProgressBar(progBar, i)

    }

    resultList <- list(summaryStatsDf=summaryStatsDf, densityDf=densityDf, simParamsDf=simParamsDf, errorDf=errorDf)

    if(writeFile){
        temp <- paste(nGens, nExamples, nLearners, rBias, cGroup, sep='-')
        body <- paste0(format(Sys.time(), "%Y-%m-%d-%H%M%S"), '-', round(runif(1,100,999),digits=0))
        resultsFile <- paste0(fPrefix, 'Run-', temp, '-', body, '.xz')
        paramsFile <- paste0(fPrefix, 'Run-', temp, '-', body, '.params')
        save(resultList, file=resultsFile, compress="xz")
        write.table(resultList[[3]], file = paramsFile)
      }

return(resultList)
}

